unit DUTUtilsTrash;
(*
   Daniel U. Thibault
   <D.U.Thibault@Bigfoot.com>
   12 September 1999
   revised 14 December 1999,
            2 April 2002 - Fixed serious sequence number bug
            24 May 2002 - Fixed Win 2k lower case drive letter bug
            02 March 2009 - In SIDtoStr, replaced '%d' format with '%u' format.
                            In GetInfoRootPath, NERR_DCNotFound no longer raises an (unjustified) exception;
                            also tightened type of cbSID from Integer to DWord.
                            In DUTGetFileSize, replaced EIntOverflow.Create(SIntOverflow)
                            with EIntOverflow.CreateRes(65532) (according to richedit_DRC.rc)
                            Seems D7's SysUtils no longer exposes it (although it still uses it)?
                            In GetTrashContents, moved the GetInfoRootPath call inside the if not Assigned(Infos[Drive]).

    1999-2009 Daniel U. Thibault.  All Rights Reserved.

   This is freeware.  No legally binding promises.  If you use
   any part of this code, you must include an acknowledgement in the
   finished code or executable (in the About box and/or the Help file).

   Extracted from my DUTUtilsFiles unit, this unit presents in
   stand-alone form the Trash (Recycle Bin) utilities.
   Several other utility functions were copied from other utility
   units of mine so as to remove any dependencies.
*)
interface

uses
   Classes,  { TList, TFileStream, etc. }
   Windows;  { THandle, TWin32FindData, MAX_PATH, etc. }

const
   sBackslash = '\';

type
   TDriveLetter = 'A'..'Z';

procedure OLECheckBoolean(Result : Boolean);
{
   Transforms a boolean Result into an HResult-like
   value and passes it on to OLECheck.
}

{ #################### Integer Flag Manipulation #################### }

function FlagIncluded(const Flags, Flag : Integer) : Boolean;
{
   Returns True if the bit pattern Flag is included in Flags.
}

{ #################### System Info #################### }

function IsWindowsNT : Boolean;
{
   Returns True if the OS is Windows NT
}

function DUTGetUserName : String;
{
   Obtains the User Name.
   Returns the '.Default' string if no "user" has logged on.
}

{ #################### Menu Manipulation #################### }

function GetMenuItemInfoByCaption(Menu : HMENU; Caption : String; var MII : TMenuItemInfo) : Boolean;
{
   Fills the TMenuItemInfo for a menu item by its caption.
   The search is not case-sensitive and is recursive.
   Returns False if no menu item had the caption sought.
   The MII.dwTypeData member should not be used since
   it'll point to a freed String (it matches Caption anyway).
}

{ #################### File Utility Functions #################### }

function HasTrailer(const FileName : String) : Boolean;
{
   Returns True if FileName ends with a '\', False otherwise.
}

function RemoveTrailer(var FileName : String) : Boolean;
{
   RemoveTrailer removes the closing '\' (if any) from FileName.
   Returns True if there was one, False otherwise.
}

function ExtractFileTitle(const FileName : String) : String;
{
   ExtractFileTitle extracts the title part of the given
   filename.  The resulting string is the rightmost characters of FileName,
   starting with the first character after the colon or backslash that
   separates the path information from the name and stopping short of the
   period and extension.
   The resulting string is empty if FileName contains no title.
}

function DUTGetFileSize(const FileName: string) : Integer;
{
   Returns the FileName's size, in bytes.
   -1 if it does not exist (or is inaccessible), 0 if empty or a directory.
   Huge files (whose size exceeds $7FFF FFFF bytes)
   will raise an EIntOverflow exception.
}

{ #################### WinNT Security ID (SID) Functions #################### }

function SIDToStr(SID : PSID) : String;
{
   Converts the SID into its string representation.
   An exception is raised if the SID is invalid.
   The string will be of the form
   'S-' + SID_REVISION + '-' + SID_IDENTIFIER_AUTHORITY + ['-' + SIDSubAuthority]*
   SID_REVISION is a single byte
   SID_IDENTIFIER_AUTHORITY is 0x000000000000
   SIDSubAuthorities are DWords

   We need this function because ConvertSIDToStringSID is only available in Win2000 (NT 5).

   Based on code supplied by
   Colin Wilson <colin@wilsonc.demon.co.uk>
   15 October 1999
}

{ #################### Trash (Recycle Bin) Functions #################### }

function GetTrashContents(DeletionDateTimes, FileSizes : TList) : TStringList;
{
   Returns a TStringList loaded with the fully qualified names of the files
   currently in the Trash (a.k.a Recycle Bin).
   The TStringList could be empty.
   It is the responsability of the calling application to Free the TStringList.
   If either or both of DeletionDateTimes and FileSizes are non-Nil, they will be
   loaded (in the same sequence as the TStringList) with the files' deletion
   times (as pointers to TDateTime objects) and sizes (as pointers to Integers).
   It is the responsability of the calling application to Free the pointed
   TDateTime and/or Integer objects.
}

function RestoreFromTrash(FileName : String) : Boolean;
{
   Attempts to Restore the fully qualified FileName from the Trash (a.k.a Recycle Bin).
   Returns False if the FileName does not exist in the Trash.
}

function DeleteFromTrash(FileName : String) : Boolean;
{
   Attempts to Delete the fully qualified FileName from the Trash (a.k.a Recycle Bin).
   Returns False if the FileName does not exist in the Trash.
}

function EmptyTrash : Boolean;
{
   Empties the Trash (a.k.a Recycle Bin).
   Returns False if the trash was already empty.
}

implementation

uses
//Dialogs, { MessageDlg }
   SysUtils, { LastDelimiter, ExtractFileName, etc. }
   ShlObj,   { CSIDL_, IShellFolder, PItemIDList, etc. }
   Forms,    { TApplication }
   COMObj,   { OLECheck }
   ActiveX;  { IMalloc }

const
   //Strings that may need to be localised
   RecycledDirStr   = 'Recycled';
   RecycledDirNTStr = 'Recycler'; //An SID (Security ID) subdir also appends itself...
// DeleteCmdStr     = '&Delete';
   RestoreCmdStr    = 'R&estore';
   EmptyTrashCmdStr = 'Empty Recycle &Bin';

   //Language independent string
   CMDSTR_DELETE    = 'delete';

   //WinNT declarations from WinNT.h
   SID_REVISION                    =    1; // Current revision level
   //from LMErr.h
   NERR_SUCCESS                    =    0;
// NERR_BASE                       = 2100;
// NERR_UnknownDevDir              = 2116; // (NERR_BASE +  16)
// NERR_RedirectedPath             = 2117; // (NERR_BASE +  17)
// NERR_DuplicateShare             = 2118; // (NERR_BASE +  18)
// NERR_BufTooSmall                = 2123; // (NERR_BASE +  23)
// NERR_NetNameNotFound            = 2310; // (NERR_BASE + 210)
// NERR_InvalidComputer            = 2351; // (NERR_BASE + 251)
   NERR_DCNotFound                 = 2453; // (NERR_BASE + 353)

procedure OLECheckBoolean(Result : Boolean);
{
   Transforms a boolean Result into an HResult-like
   value and passes it on to OLECheck.
}
begin
   OLECheck(-Ord(not Result));
end; { OLECheckBoolean }

{ #################### Integer Flag Manipulation #################### }

function FlagIncluded(const Flags, Flag : Integer) : Boolean;
{
   Returns True if the bit pattern Flag is included in Flags.
}
begin
   Result := Flag = (Flags and Flag);
end; { FlagIncluded }

{ #################### System Info #################### }

function IsWindowsNT : Boolean;
{
   Returns True if the OS is Windows NT
}
var
   VersionInfo : TOSVersionInfo;
begin
   // set the size of the record
   VersionInfo.dwOSVersionInfoSize := SizeOf(TOSVersionInfo);
   if {Windows.}GetVersionEx(VersionInfo) then begin
      Result := VersionInfo.dwPlatformId = VER_PLATFORM_WIN32_NT;
      {
      dwPlatformID could also be
      VER_PLATFORM_WIN32s or     //Win32s
      VER_PLATFORM_WIN32_WINDOWS //Windows 95
      }
   end else begin
      raise Exception.Create('IsWindowsNT: Unable to GetVersionEx');
   end; { else }
end; { IsWindowsNT }

function DUTGetUserName : String;
{
   Obtains the User Name.
   Returns the '.Default' string if no "user" has logged on.
}
var
   Size : DWORD;
begin
   Size := MAX_PATH;
   Result := StringOfChar(#0, Size);
   if Windows.GetUserName(PChar(Result), Size) then begin
      SetLength(Result, Size - 1);
   end else begin
      Result := '.Default';
   end; { else }
end; { DUTGetUserName }

{ #################### Menu Manipulation #################### }

function GetMenuItemInfoByCaption(Menu : HMENU; Caption : String; var MII : TMenuItemInfo) : Boolean;
{
   Fills the TMenuItemInfo for a menu item by its caption.
   The search is not case-sensitive and is recursive.
   Returns False if no menu item had the caption sought.
   The MII.dwTypeData member should not be used since
   it'll point to a freed String (it matches Caption anyway).
}
var
   i        : Integer;
   ACaption : String;
begin
   Result := False;
   for i := 0 to (GetMenuItemCount(Menu) - 1) do begin
      FillChar(MII, SizeOf(TMenuItemInfo), #0);
      MII.cbSize := SizeOf(TMenuItemInfo);
      //Retrieve everything
      MII.fMask := MIIM_CHECKMARKS or MIIM_DATA or MIIM_ID or MIIM_STATE or MIIM_SUBMENU or MIIM_TYPE;
      MII.cch := 255;
      ACaption := StringOfChar(#0, MII.cch);
      MII.dwTypeData := PChar(ACaption);
      if not GetMenuItemInfo(Menu, i, True, MII) then raise Exception.Create('GetMenuItemInfoByCaption: Unable to GetMenuItemInfo for item number ' + IntToStr(i));
      //Ignore Separators and Bitmaps
      if (MII.fType = MFT_STRING) and
         (0 = ANSICompareText(Caption, String(MII.dwTypeData)))
      then begin
         Result := True;
      end else if MII.hSubMenu <> 0 then begin
         //Recurse
         Result := GetMenuItemInfoByCaption(MII.hSubMenu, Caption, MII);
      end; { else if }
      if Result then Exit;
   end; { for }
end; { GetMenuItemInfoByCaption }

{ #################### File Utility Functions #################### }

function HasTrailer(const FileName : String) : Boolean;
{
   Returns True if FileName ends with a '\', False otherwise.
}
begin
   Result := (FileName <> '') and (FileName[Length(FileName)] = sBackslash);
end; { HasTrailer }

function RemoveTrailer(var FileName : String) : Boolean;
{
   RemoveTrailer removes the closing '\' (if any) from FileName.
   Returns True if there was one, False otherwise.
}
begin
   Result := HasTrailer(FileName);
   if Result then FileName := Copy(FileName, 1, Length(FileName) - 1);
end; { RemoveTrailer }

function ExtractFileTitle(const FileName : String) : String;
{
   ExtractFileTitle extracts the title part of the given
   filename.  The resulting string is the rightmost characters of FileName,
   starting with the first character after the colon or backslash that
   separates the path information from the name and stopping short of the
   period and extension.
   The resulting string is empty if FileName contains no title.
}
var
   i, j : Integer;
begin
   i := LastDelimiter('.', FileName);
   j := LastDelimiter('\:', FileName);
   if (i > j) then
      Result := Copy(FileName, j + 1, i - (j + 1))
   else if (j > 0) then
      Result := Copy(FileName, j + 1, MaxInt)
   else
      Result := FileName;
end; { ExtractFileTitle }

function DUTGetFileSize(const FileName: string) : Integer;
{
   Returns the FileName's size, in bytes.
   -1 if it does not exist (or is inaccessible), 0 if empty or a directory.
   Huge files (whose size exceeds $7FFF FFFF bytes)
   will raise an EIntOverflow exception.
}
var
   Handle      : THandle;
   FindData    : TWin32FindData;
   WorkingName : String;
begin
   Result := -1; //File does not exist
   WorkingName := FileName;
   RemoveTrailer(WorkingName);
   if WorkingName = '' then Exit;
   Handle := Windows.FindFirstFile(PChar(WorkingName), FindData);
   if Handle = INVALID_HANDLE_VALUE then Exit;
   Windows.FindClose(Handle);
   Result := 0; //File is a directory
   if FlagIncluded(FindData.dwFileAttributes, FILE_ATTRIBUTE_DIRECTORY) then Exit;
// True Size == (FindData.nFileSizeHigh * MAXDWORD { $FFFFFFFF } ) + FindData.nFileSizeLow;
   //Delphi LongInt is limited to a maximum of $7FFFFFFF
   Result := FindData.nFileSizeLow;
   // Here we use the Delphi RTL-supplied localized error string
   if Result < 0 then raise EIntOverflow.CreateRes(65532); //See richedit_DRC.rc
end; { DUTGetFileSize }

{ #################### WinNT Security ID (SID) Functions #################### }

function SIDToStr(SID : PSID) : String;
{
   Converts the SID into its string representation.
   An exception is raised if the SID is invalid.
   The string will be of the form
   'S-' + SID_REVISION + '-' + SID_IDENTIFIER_AUTHORITY + ['-' + SIDSubAuthority]*
   SID_REVISION is a single byte
   SID_IDENTIFIER_AUTHORITY is 0x000000000000
   SIDSubAuthorities are DWords

   We need this function because ConvertSIDToStringSID is only available in Win2000 (NT 5).

   Based on code supplied by
   Colin Wilson <colin@wilsonc.demon.co.uk>
   15 October 1999
}
var
   pSIA             : PSIDIdentifierAuthority;
   dwSubAuthorities,
   dwSIDRev,
   dwCounter        : DWORD;
begin
   dwSIDRev := SID_REVISION;
   //Test if SID passed in is valid
   if not IsValidSID(SID) then raise Exception.Create('SIDtoStr: Invalid SID');

   pSIA := GetSIDIdentifierAuthority(SID);
   dwSubAuthorities := GetSIDSubAuthorityCount(SID)^;
   //Prepare S-SID_REVISION-
// Result := Format('S-%d-', [dwSIDRev]);
   Result := Format('S-%u-', [dwSIDRev]);
   //Prepare SID.IdentifierAuthority
   if (pSIA^.Value[0] <> 0) or (pSIA^.Value[1] <> 0) then begin
      { Note: This hex format is undocumented }
//    Result := Result + Format('0x%02x%02x%02x%02x%02x%02x', [
      Result := Result + Format('0x%2x%2x%2x%2x%2x%2x', [
                                 pSIA^.Value[0],
                                 pSIA^.Value[1],
                                 pSIA^.Value[2],
                                 pSIA^.Value[3],
                                 pSIA^.Value[4],
                                 pSIA^.Value[5] ]);
   end else begin
      //Value[0..1] are ignored; should be zero
//    Result := Result + Format('%d', [
      Result := Result + Format('%u', [
                                 DWORD(pSIA^.Value[5]) +
                                 DWORD(pSIA^.Value[4] shl  8) +
                                 DWORD(pSIA^.Value[3] shl 16) +
                                 DWORD(pSIA^.Value[2] shl 24)   ]);
   end; { else }
   //Loop through SID.SubAuthority array
   for dwCounter := 0 to (dwSubAuthorities - 1) do begin
//    Result := Result + Format('-%d', [GetSIDSubAuthority(SID, dwCounter)^]);
      Result := Result + Format('-%u', [GetSIDSubAuthority(SID, dwCounter)^]);
   end; { for }
end; { SIDToStr }

{ #################### Trash (Recycle Bin) Functions #################### }

function GetInfoRootPath(Drive : String) : String;
{
   Private function used by GetTrashContents and RestoreFromTrash.
   Returns the path to the Drive's recycled/recycler (sub)dir containing
   the Info or Info2 file.
   The concluding backslash is included.
}
var
   NTNetGetDCName     : function(Server, Domain : pWideChar; var DC : pWideChar) : DWORD {NET_API_STATUS}; stdcall;
   NTNetAPIBufferFree : function(lpBuffer : Pointer) : DWORD {NET_API_STATUS}; stdcall;

   PrimaryDomain : String;
   ErrMode       : Word;
   LibHandle     : THandle;
   pDomain       : PWideChar;
   cbSID         : DWord; //Integer;
   SID           : PSID;
   peUse         : SID_NAME_USE;
   DomainName    : packed array [1..MAX_PATH] of Char;
   cbDomainName  : DWord;
   i             : Integer;
begin
   Result := Drive + ':\' + RecycledDirStr + sBackslash;
   if not IsWindowsNT then Exit;

(*
   The location of the Info[2] file is different under WinNT:
   It appears as <Drive>:\Recycler\<some weird dir name>\Info
   According to Microsoft Knowledge Base Article Q171694, this
   weird dir name is a Security Identifier (SID); each logged-on
   user will thus see his own set of recycled files.
*)
   //Identify the WinNT intermediate directory
   {
   To obtain the correct SID, we must know the Primary Domain Controller's
   name (the domain the user logged onto).
   }
   PrimaryDomain := '';
   //We must dynamically bind to NetAPI32.dll since NetGetDCName and
   //NetAPIBufferFree are not available under Win95
   {
   Well, actually, NetGetDCName exists in 16-bit NetAPI.dll but its
   syntax is different: instead of passing a variable that receives
   a buffer address that must later be freed by NetAPIBufferFree, you
   pass it a buffer and its length.  NetGetDCName then writes the
   primary domain controller's name to the buffer.
   The return codes are the same.
   Example:
   cbDomain := MAX_PATH;
   GetMem(pDomain, cbDomain);
   if NERR_SUCCESS <> NetGetDCName(Nil, Nil, pDomain, cbDomain) then
   }
   ErrMode := SetErrorMode(SEM_NOOPENFILEERRORBOX);
   LibHandle := LoadLibrary('NetAPI32.dll');
   SetErrorMode(ErrMode);
   if LibHandle = 0 then raise Exception.Create('Unable to load library NetAPI32.dll');
   try { LibHandle }
      @NTNetGetDCName     := GetProcAddress(LibHandle, 'NetGetDCName');
      @NTNetAPIBufferFree := GetProcAddress(LibHandle, 'NetApiBufferFree');
      SetLastError(NTNetGetDCName(Nil, Nil, pDomain));
      i := GetLastError; // Must store this, as SysErrorMessage() will reset the last error to zero
      if NERR_SUCCESS <> i then begin
         if NERR_DCNotFound = i then begin
            PrimaryDomain := ''; //If not connected, just ignore
         end else begin
            //I'm not sure SysErrorMessage will work since the NERR_ codes are in the 2100..2999 range
            raise Exception.CreateFmt('%s (Error code %d)', [SysErrorMessage(i), i]);
         end; //if
      end else begin
         try { pDomain }
            //MSDN Online says the PrimaryDomain thus retrieved will be
            //prefixed by '\\' but that does not seem to be the case
            PrimaryDomain := WideCharToString(pDomain);
//MessageDlg('GetInfoRootPath: PrimaryDomain is  ' + PrimaryDomain + ' ', mtCustom, [mbOK], 0);
         finally
            NTNetAPIBufferFree(pDomain);
         end; { try }
      end; //if
   finally
      FreeLibrary(LibHandle);
   end; { try }

   //Allocate SID
   cbSID := 1024;
   cbDomainName := MAX_PATH;
   GetMem(SID, cbSID);
   try { SID }
      //LookupAccountName isn't supposed to be available under Win95,
      //according to Delphi 3's Win32.hlp.  However, my Win95 system's
      //AdvAPI32.dll does include it, so...
      //if not LookupAccountName(PChar(DUTGetSecurityProvider), PChar(DUTGetUserName), SID, cbSID, @DomainName, cbDomainName, peUse) then begin
//MessageDlg('GetInfoRootPath: DUTGetUserName is  ' + DUTGetUserName + ' ', mtCustom, [mbOK], 0);
      if not LookupAccountName(PChar(PrimaryDomain),
                               PChar(DUTGetUserName),
                               SID,
                               cbSID,
                               @DomainName,
                               cbDomainName,
                               peUse) then
         raise Exception.Create('Unable to obtain SID');
      //peUse ought to be SIDTypeUser
      Result := Drive + ':\' + RecycledDirNTStr + sBackslash + SIDToStr(SID) + sBackslash;
//MessageDlg('GetInfoRootPath: Result is  ' + Result + ' ', mtCustom, [mbOK], 0);
   finally
      FreeMem(SID); //FreeSID(SID);
   end; { try }
end; { GetInfoRootPath }

function GetTrashContents(DeletionDateTimes, FileSizes : TList) : TStringList;
{
   Returns a TStringList loaded with the fully qualified names of the files
   currently in the Trash (a.k.a Recycle Bin).
   The TStringList could be empty.
   It is the responsability of the calling application to Free the TStringList.
   If either or both of DeletionDateTimes and FileSizes are non-Nil, they will be
   loaded (in the same sequence as the TStringList) with the files' deletion
   times (as pointers to TDateTime objects) and sizes (as pointers to Integers).
   It is the responsability of the calling application to Free the pointed
   TDateTime and/or Integer objects.
}
var
   SM           : IMalloc;
   SF,
   TF           : IShellFolder;
   ItemPIDL,
   TrashPIDL    : PItemIDList;
   EL           : IEnumIDList;
   Fetched      : ULONG;
   Drive        : TDriveLetter;
   NumFile,
   NumSought    : Integer;
   InfosPos     : LongInt;
   FilePath,
   InfoPath     : packed array [1..MAX_PATH] of Char; //Must be precisely MAX_PATH bytes
   InfoRootPath : String;
   Infos        : array[TDriveLetter] of TFileStream;
   InfoNum      : array[TDriveLetter] of Integer;
   DeletedOn,
   DeletedOn2        : TFileTime;
   FileSize,
   DOSDelDateTime    : Integer;
   pDeletionDateTime : ^TDateTime;
   pFileSize         : ^Integer;
begin
   Result := TStringList.Create;
   try { Result }
      for Drive := Low(TDriveLetter) to High(TDriveLetter) do Infos[Drive] := Nil;
      //We'll need IMalloc to free the returned PIDLs
      OLECheck(SHGetMalloc(SM));
      //Obtain IShellFolder for desktop
      OLECheck(SHGetDesktopFolder(SF));
      //Obtain Trash
      OLECheck(SHGetSpecialFolderLocation(Application.Handle, CSIDL_BITBUCKET, TrashPIDL));
      try { TrashPIDL }
         //Bind to the Trash
         OLECheck(SF.BindToObject(TrashPIDL, Nil, IID_IShellFolder, Pointer(TF)));
      finally
         //We no longer need the TrashPIDL
         SM.Free(TrashPIDL);
      end; { try }
      //Obtain IEnumIDList of Trash contents
      OLECheck(TF.EnumObjects(Application.Handle, SHCONTF_NONFOLDERS or SHCONTF_INCLUDEHIDDEN, EL));
      //Fetch first item in Trash
      //Note that S_FALSE is NOT considered an error by OLECheck
      OLECheck(EL.Next(1, ItemPIDL, Fetched));
      try { Infos[] }
         while Fetched > 0 do begin
            try { ItemPIDL }
               {
               At this point, calling SHGetPathFromIDList(ItemPIDL, ...) returns the
               true file title ('D'<OriginalDriveLetter><Number>.<extension>)
               but always returns the path as 'C:\Windows\Desktop\' since the
               ItemPIDL isn't relative to the Desktop (as SHGetPathFromIDList
               expects) but rather relative to the Trash.
               We'll use the file title information to locate the proper
               'Recycled' directory and read it's Info (or Info2) file...
               }
               OLECheckBoolean(SHGetPathFromIDList(ItemPIDL, @InfoPath));
            finally
               //We no longer need this PIDL
               SM.Free(ItemPIDL);
            end; { try }

            //InfoPath is 'C:\Windows\Desktop\D'<OriginalDriveLetter><Number>.<extension>
            //         or 'C:\Windows\Desktop\D'<OriginalDriveLetter><Number>
            //            'C:\Documents and Settings\<username>\Desktop\D'<OriginalDriveLetter><Number>[.<extension>] (Win 2000)
            //The item may be a directory.
//MessageDlg('GetTrashContents: InfoPath for the first deleted file is  ' + String(PChar(@InfoPath)) + ' ', mtCustom, [mbOK], 0);
            //Extract drive letter from trash file title
            //Win 2k actually stores the drive letter in lower case
            Drive := UpperCase(ExtractFileName(String(PChar(@InfoPath))))[2];
//MessageDlg('GetTrashContents: Drive is  ' + Drive + ' ', mtCustom, [mbOK], 0);
            //Extract sequence number from trash file title
            NumSought := StrToInt(Copy(ExtractFileTitle(String(PChar(@InfoPath))), 3, MAX_PATH));
//MessageDlg('GetTrashContents: NumSought is  ' + IntToStr(NumSought) + ' ', mtCustom, [mbOK], 0);

            //If we haven't opened the corresponding drive's Info file yet, do so
            if not Assigned(Infos[Drive]) then begin
               //Obtain the path to the Info[2] file
               InfoRootPath := GetInfoRootPath(Drive);
//MessageDlg('GetTrashContents: InfoRootPath is  ' + InfoRootPath + ' ', mtCustom, [mbOK], 0);
               try { Info2 vs Info }
                  Infos[Drive] := TFileStream.Create(InfoRootPath + 'Info2', fmOpenRead or fmShareDenyNone);
               except
                  Infos[Drive] := TFileStream.Create(InfoRootPath + 'Info',  fmOpenRead or fmShareDenyNone);
               end; { try }
               //Skip 20-byte header
               Infos[Drive].Seek(20, soFromBeginning);
            end; { then }
            //Look ahead to the next record's FileSequenceNumber
            InfosPos := Infos[Drive].Position;
            Infos[Drive].Seek(MAX_PATH, soFromCurrent);
            Infos[Drive].Read(NumFile, SizeOf(NumFile));
            InfoNum[Drive] := NumFile;
            //Rewind
            Infos[Drive].Position := InfosPos;

            //Read the file info
            //The EL returns the files in the same sequence as they appear in
            //the Recycled\Info file; however, there may be gaps in the latter.
            //Skip gaps
            while InfoNum[Drive] < NumSought do begin
               //Skip a record
               Infos[Drive].Seek(MAX_PATH*(1 + 2*Ord(IsWindowsNT)) + 20, soFromCurrent);
               //Look ahead to the next record's FileSequenceNumber
               InfosPos := Infos[Drive].Position;
               Infos[Drive].Seek(MAX_PATH, soFromCurrent);
               Infos[Drive].Read(NumFile, SizeOf(NumFile));
               InfoNum[Drive] := NumFile;
               //Rewind
               Infos[Drive].Position := InfosPos;
            end; { while }

            //Read the file's fully qualified original path and name
            Infos[Drive].Read(FilePath, MAX_PATH);
            //The String(PChar()) typecast sets the Length of the string to its correct value
            Result.Add(String(PChar(@FilePath)));

            //Skip 8 bytes:
            // FileSequenceNumber : Integer;
            // DriveLetter        : Integer; { 'A' = 0, 'B' = 1... }
            Infos[Drive].Seek(8, soFromCurrent);

            //Get deletion time
            //DeletedOn : TFileTime is the UTC time of deletion
            Infos[Drive].Read(DeletedOn, SizeOf(TDateTime));
            if Assigned(DeletionDateTimes) then begin
               //Convert from UTC to local time
               DeletedOn2 := DeletedOn;
               OLECheckBoolean(FileTimeToLocalFileTime(DeletedOn2, DeletedOn));
               //Convert to DOS date-time value
               OLECheckBoolean(FileTimeToDosDateTime(DeletedOn, LongRec(DOSDelDateTime).Hi, LongRec(DOSDelDateTime).Lo));
               GetMem(pDeletionDateTime, SizeOf(TDateTime));
               try { pDeletionDateTime }
                  //Convert to TDateTime
                  pDeletionDateTime^ := FileDateToDateTime(DOSDelDateTime);
                  DeletionDateTimes.Add(pDeletionDateTime);
               except
                  FreeMem(pDeletionDateTime);
                  raise;
               end; { try }
            end; { then }

            //Get file size
            {
            This obtains the file's footprint size; that is,
            the number of bytes it occupies on the drive
            (a whole multiple of the drive's cluster size)
            }
            //We need to do the Read just to advance the file cursor
            Infos[Drive].Read(FileSize, SizeOf(Integer));
            {
            This obtains the file's true size; that is,
            the number of bytes it contains
            }
            FileSize := DUTGetFileSize(Drive + ':\' + RecycledDirStr + sBackslash + ExtractFileName(String(PChar(@InfoPath))));
            if Assigned(FileSizes) then begin
               GetMem(pFileSize, SizeOf(Integer));
               try { pFileSize }
                  pFileSize^ := FileSize;
                  FileSizes.Add(pFileSize);
               except
                  FreeMem(pFileSize);
                  raise;
               end; { try }
            end; { then }

            //Skip Unicode file name
            if IsWindowsNT then Infos[Drive].Seek(MAX_PATH*2, soFromCurrent);

            //Next item
            OLECheck(EL.Next(1, ItemPIDL, Fetched));
         end; { while }
      finally
         //Cleanup --close all Recycled\Info files
         for Drive := Low(TDriveLetter) to High(TDriveLetter) do Infos[Drive].Free;
      end; { try }
   except
      Result.Free;
//    Result := Nil;
      raise;
   end; { try }
end; { GetTrashContents }

function RestoreFromTrash(FileName : String) : Boolean;
{
   Attempts to Restore the fully qualified FileName from the Trash (a.k.a Recycle Bin).
   Returns False if the FileName does not exist in the Trash.
}
const
   CmdID_First = 0;
var
   SM           : IMalloc;
   TF,
   SF           : IShellFolder;
   ItemPIDL,
   TrashPIDL    : PItemIDList;
   EL           : IEnumIDList;
   Fetched      : ULONG;
   InfoNum,
   NumSought    : Integer;
   InfosPos     : LongInt;
   InfoPath,
   FilePath     : packed array [1..MAX_PATH] of Char; //Must be precisely MAX_PATH bytes
   InfoRootPath : String;
   Info         : TFileStream;
   ContextMenu  : IContextMenu;
   Popup        : HMenu;
   ICI          : TCMInvokeCommandInfo;
   MII          : TMenuItemInfo;
   hObject      : THandle;
begin
   Result := False;
   if FileName = '' then Exit;
   OLECheck(SHGetMalloc(SM));;
   //Obtain IShellFolder for desktop
   OLECheck(SHGetDesktopFolder(SF));
   //Obtain Trash
   OLECheck(SHGetSpecialFolderLocation(Application.Handle, CSIDL_BITBUCKET, TrashPIDL));
   try { TrashPIDL }
      //Obtain the path to the Info[2] file
      InfoRootPath := GetInfoRootPath(FileName[1]);

      //Open the corresponding Recycled\Info file
      try { Info2 vs Info }
         Info := TFileStream.Create(InfoRootPath + 'Info2', fmOpenRead or fmShareDenyNone);
      except
         Info := TFileStream.Create(InfoRootPath + 'Info',  fmOpenRead or fmShareDenyNone);
      end; { try }

      try { Info }
         //Skip 20-byte header
         Info.Seek(20, soFromBeginning);

         //Bind to the Trash
         OLECheck(SF.BindToObject(TrashPIDL, Nil, IID_IShellFolder, Pointer(TF)));
         //Obtain IEnumIDList of Trash contents
         OLECheck(TF.EnumObjects(Application.Handle, SHCONTF_NONFOLDERS or SHCONTF_INCLUDEHIDDEN, EL));
         //Fetch first item in Trash
         //Note that S_FALSE is NOT considered an error by OLECheck
         OLECheck(EL.Next(1, ItemPIDL, Fetched));
         try { ItemPIDL }
            while Fetched > 0 do begin
               OLECheckBoolean(SHGetPathFromIDList(ItemPIDL, @InfoPath));
               //InfoPath is 'C:\Windows\Desktop\D'<OriginalDriveLetter><Number>.<extension>
               //Extract drive letter from trash file title
               //If not the FileName's drive, move on to next item in Trash
               if 0 = ANSICompareText(FileName[1], ExtractFileName(String(PChar(@InfoPath)))[2]) then begin
                  //Extract sequence number from trash file title
                  NumSought := StrToInt(Copy(ExtractFileTitle(String(PChar(@InfoPath))), 3, MAX_PATH));

                  //Look ahead to the record's FileSequenceNumber
                  InfosPos := Info.Position;
                  Info.Seek(MAX_PATH, soFromCurrent);
                  Info.Read(InfoNum, SizeOf(InfoNum));
                  //Rewind
                  Info.Position := InfosPos;

                  //Read the file info
                  //The EL returns the files in the same sequence as they appear in
                  //the Recycled\Info file; however, there may be gaps in the latter.
                  //Skip gaps
                  while InfoNum < NumSought do begin
                     //Skip a record
                     Info.Seek(MAX_PATH*(1 + 2*Ord(IsWindowsNT)) + 20, soFromCurrent);
                     //Look ahead to the next record's FileSequenceNumber
                     InfosPos := Info.Position;
                     Info.Seek(MAX_PATH, soFromCurrent);
                     Info.Read(InfoNum, SizeOf(InfoNum));
                     //Rewind
                     Info.Position := InfosPos;
                  end; { while }

                  //Read the file's original fully qualified name
                  Info.Read(FilePath, MAX_PATH);
                  //Skip the remaining 20 bytes
                  Info.Seek(20, soFromCurrent);
                  //Skip Unicode file name
                  if IsWindowsNT then Info.Seek(MAX_PATH*2, soFromCurrent);

                  //The String(PChar()) typecast sets the Length of the string to its correct value
                  if 0 = ANSICompareText(FileName, String(PChar(@FilePath))) then break;
               end; { then }
               //Next item in Trash
               SM.Free(ItemPIDL); ItemPIDL := Nil;
               OLECheck(EL.Next(1, ItemPIDL, Fetched));
            end; { while }
            //Did we find FileName?
            //Note that this failure mode is not an Exception
            if Fetched = 0 then Exit;

            //We found the FileName
            //Obtain IContextMenu
            OLECheck(TF.GetUIObjectOf(Application.Handle, 1, ItemPIDL, IID_IContextMenu, Nil, Pointer(ContextMenu)));
         finally
            //PIDL no longer needed
            //However, it may have already been Freed in two ways:
            // 1) We did not find the FileName; or
            // 2) EL.Next failed
            if Assigned(ItemPIDL) then SM.Free(ItemPIDL);
         end; { try }
      finally
         Info.Free;
      end; { try }

      //Populate menu
      Popup := CreatePopupMenu;
      OLECheckBoolean(Popup <> 0);
      try { Popup }
         OLECheck(ContextMenu.QueryContextMenu(Popup, 0 {insertion index}, CmdID_First, $7FFF {CmdID_Last}, CMF_EXPLORE));
         //Identify the "Restore" command (unfortunately, the Shell supplies no Verb!)
         OLECheckBoolean(GetMenuItemInfoByCaption(Popup, RestoreCmdStr, MII));
         //Invoke the command
         FillChar(ICI, SizeOf(TCMInvokeCommandInfo), #0);
         with ICI do begin
            cbSize := SizeOf(TCMInvokeCommandInfo);
            //Does NOT prevent confirmation prompt (for read-only files and such)
//          fMask := CMIC_MASK_FLAG_NO_UI;
            hWnd := Application.Handle;
            nShow := SW_SHOWNORMAL;
            lpVerb := PChar(MII.wID - CmdID_First);
         end; { with }
         OLECheck(ContextMenu.InvokeCommand(ICI));
         Result := True;

         //It seems the shell dawdles a bit before updating the Info files
         SHChangeNotify(SHCNE_UPDATEDIR, SHCNF_FLUSH or SHCNF_IDLIST, TrashPIDL, Nil);
         hObject := FindFirstChangeNotification(PChar(FileName[1] + ':\' + RecycledDirStr), False, FILE_NOTIFY_CHANGE_LAST_WRITE);
         OLECheckBoolean(INVALID_HANDLE_VALUE <> hObject);
         try { hObject }
            //We do not want to throw an exception if the user is faced with a confirmation prompt
//          OLECheckBoolean(WAIT_OBJECT_0 = WaitForSingleObject(hObject, 500 { ms }));
            WaitForSingleObject(hObject, 500 { ms });
         finally
            OLECheckBoolean(FindCloseChangeNotification(hObject));
         end; { try }
      finally
         OLECheckBoolean(DestroyMenu(Popup));
      end; { try }
   finally
      SM.Free(TrashPIDL);
   end; { try }
end; { RestoreFromTrash }

function DeleteFromTrash(FileName : String) : Boolean;
{
   Attempts to Delete the fully qualified FileName from the Trash (a.k.a Recycle Bin).
   Returns False if the FileName does not exist in the Trash.
}
const
   CmdID_First = 0;
var
   SM           : IMalloc;
   TF,
   SF           : IShellFolder;
   ItemPIDL,
   TrashPIDL    : PItemIDList;
   EL           : IEnumIDList;
   Fetched      : ULONG;
   InfoNum,
   NumSought    : Integer;
   InfosPos     : LongInt;
   InfoPath,
   FilePath     : packed array [1..MAX_PATH] of Char; //Must be precisely MAX_PATH bytes
   InfoRootPath : String;
   Info         : TFileStream;
   ContextMenu  : IContextMenu;
   Popup        : HMenu;
   ICI          : TCMInvokeCommandInfo;
   { This is one departure from the RestoreFromTrash code }
// MII          : TMenuItemInfo;
   hObject      : THandle;
begin
   Result := False;
   if FileName = '' then Exit;
   OLECheck(SHGetMalloc(SM));;
   //Obtain IShellFolder for desktop
   OLECheck(SHGetDesktopFolder(SF));
   //Obtain Trash
   OLECheck(SHGetSpecialFolderLocation(Application.Handle, CSIDL_BITBUCKET, TrashPIDL));
   try { TrashPIDL }
      //Obtain the path to the Info[2] file
      InfoRootPath := GetInfoRootPath(FileName[1]);

      //Open the corresponding Recycled\Info file
      try { Info2 vs Info }
         Info := TFileStream.Create(InfoRootPath + 'Info2', fmOpenRead or fmShareDenyNone);
      except
         Info := TFileStream.Create(InfoRootPath + 'Info',  fmOpenRead or fmShareDenyNone);
      end; { try }

      try { Info }
         //Skip 20-byte header
         Info.Seek(20, soFromBeginning);

         //Bind to the Trash
         OLECheck(SF.BindToObject(TrashPIDL, Nil, IID_IShellFolder, Pointer(TF)));
         //Obtain IEnumIDList of Trash contents
         OLECheck(TF.EnumObjects(Application.Handle, SHCONTF_NONFOLDERS or SHCONTF_INCLUDEHIDDEN, EL));
         //Fetch first item in Trash
         //Note that S_FALSE is NOT considered an error by OLECheck
         OLECheck(EL.Next(1, ItemPIDL, Fetched));
         try { ItemPIDL }
            while Fetched > 0 do begin
               OLECheckBoolean(SHGetPathFromIDList(ItemPIDL, @InfoPath));
               //InfoPath is 'C:\Windows\Desktop\D'<OriginalDriveLetter><Number>.<extension>
               //Extract drive letter from trash file title
               //If not the FileName's drive, move on to next item in Trash
               if 0 = ANSICompareText(FileName[1], ExtractFileName(String(PChar(@InfoPath)))[2]) then begin
                  //Extract sequence number from trash file title
                  NumSought := StrToInt(Copy(ExtractFileTitle(String(PChar(@InfoPath))), 3, MAX_PATH));

                  //Look ahead to the next record's FileSequenceNumber
                  InfosPos := Info.Position;
                  Info.Seek(MAX_PATH, soFromCurrent);
                  Info.Read(InfoNum, SizeOf(InfoNum));
                  //Rewind
                  Info.Position := InfosPos;

                  //Read the file info
                  //The EL returns the files in the same sequence as they appear in
                  //the Recycled\Info file; however, there may be gaps in the latter.
                  //Skip gaps
                  while InfoNum < NumSought do begin
                     Info.Seek(MAX_PATH*(1 + 2*Ord(IsWindowsNT)) + 20, soFromCurrent);
                     //Look ahead to the next record's FileSequenceNumber
                     InfosPos := Info.Position;
                     Info.Seek(MAX_PATH, soFromCurrent);
                     Info.Read(InfoNum, SizeOf(InfoNum));
                     //Rewind
                     Info.Position := InfosPos;
                  end; { while }

                  //Read the file's original fully qualified name
                  Info.Read(FilePath, MAX_PATH);
                  //Skip the remaining 20 bytes
                  Info.Seek(20, soFromCurrent);
                  //Skip Unicode file name
                  if IsWindowsNT then Info.Seek(MAX_PATH*2, soFromCurrent);

                  //The String(PChar()) typecast sets the Length of the string to its correct value
                  if 0 = ANSICompareText(FileName, String(PChar(@FilePath))) then break;
               end; { then }
               //Next item in Trash
               SM.Free(ItemPIDL); ItemPIDL := Nil;
               OLECheck(EL.Next(1, ItemPIDL, Fetched));
            end; { while }
            //Did we find FileName?
            //Note that this failure mode is not an Exception
            if Fetched = 0 then Exit;

            //We found the FileName
            //Obtain IContextMenu
            OLECheck(TF.GetUIObjectOf(Application.Handle, 1, ItemPIDL, IID_IContextMenu, Nil, Pointer(ContextMenu)));
         finally
            //PIDL no longer needed
            //However, it may have already been Freed in two ways:
            // 1) We did not find the FileName; or
            // 2) EL.Next failed
            if Assigned(ItemPIDL) then SM.Free(ItemPIDL);
         end; { try }
      finally
         Info.Free;
      end; { try }

      //Populate menu
      Popup := CreatePopupMenu;
      OLECheckBoolean(Popup <> 0);
      try { Popup }
         OLECheck(ContextMenu.QueryContextMenu(Popup, 0 {insertion index}, CmdID_First, $7FFF {CmdID_Last}, CMF_EXPLORE));
         { This is another departure from the RestoreFromTrash code }
         //No need to identify the "Delete" command since the Shell does supply a Verb in this case
//       OLECheckBoolean(GetMenuItemInfoByCaption(Popup, DeleteCmdStr, MII));
         //Invoke the command
         FillChar(ICI, SizeOf(TCMInvokeCommandInfo), #0);
         with ICI do begin
            cbSize := SizeOf(TCMInvokeCommandInfo);
            //Does NOT prevent confirmation prompt (for read-only files and such)
//          fMask := CMIC_MASK_FLAG_NO_UI;
            hWnd := Application.Handle;
            nShow := SW_SHOWNORMAL;
            { This is the last departure from the RestoreFromTrash code }
//          lpVerb := PChar(MII.wID - CmdID_First);
            lpVerb := PChar(CMDSTR_DELETE);
         end; { with }
         OLECheck(ContextMenu.InvokeCommand(ICI));
         Result := True;

         //It seems the shell dawdles a bit before updating the Info files
         SHChangeNotify(SHCNE_UPDATEDIR, SHCNF_FLUSH or SHCNF_IDLIST, TrashPIDL, Nil);
         hObject := FindFirstChangeNotification(PChar(FileName[1] + ':\' + RecycledDirStr), False, FILE_NOTIFY_CHANGE_LAST_WRITE);
         OLECheckBoolean(INVALID_HANDLE_VALUE <> hObject);
         try { hObject }
            //We do not want to throw an exception if the user is faced with a confirmation prompt
//          OLECheckBoolean(WAIT_OBJECT_0 = WaitForSingleObject(hObject, 500 { ms }));
            WaitForSingleObject(hObject, 500 { ms });
         finally
            OLECheckBoolean(FindCloseChangeNotification(hObject));
         end; { try }
      finally
         OLECheckBoolean(DestroyMenu(Popup));
      end; { try }
   finally
      SM.Free(TrashPIDL);
   end; { try }
end; { DeleteFromTrash }

function EmptyTrash : Boolean;
{
   Empties the Trash (a.k.a Recycle Bin).
   Returns False if the trash was already empty.
}
const
   CmdID_First = 0;
var
   SM          : IMalloc;
   SF          : IShellFolder;
   TrashPIDL   : PItemIDList;
   ContextMenu : IContextMenu;
   Popup       : HMenu;
   ICI         : TCMInvokeCommandInfo;
   MII         : TMenuItemInfo;
begin
   Result := False;
   OLECheck(SHGetMalloc(SM));
   //Obtain IShellFolder for desktop
   OLECheck(SHGetDesktopFolder(SF));
   //Obtain Trash
   OLECheck(SHGetSpecialFolderLocation(Application.Handle, CSIDL_BITBUCKET, TrashPIDL));
   try { TrashPIDL }
      //Obtain IContextMenu
      OLECheck(SF.GetUIObjectOf(Application.Handle, 1, TrashPIDL, IID_IContextMenu, Nil, Pointer(ContextMenu)));
      //Populate menu
      Popup := CreatePopupMenu;
      OLECheckBoolean(Popup <> 0);
      try { Popup }
         OLECheck(ContextMenu.QueryContextMenu(Popup, 0 {insertion index}, CmdID_First, $7FFF {CmdID_Last}, CMF_EXPLORE));
         //Identify the "Empty Recycle Bin" command (unfortunately, the Shell supplies no Verb!)
         OLECheckBoolean(GetMenuItemInfoByCaption(Popup, EmptyTrashCmdStr, MII));
         //Enabled?
         if FlagIncluded(MII.fState, MFS_ENABLED) then begin
            //Invoke the command
            FillChar(ICI, SizeOf(TCMInvokeCommandInfo), #0);
            with ICI do begin
               cbSize := SizeOf(TCMInvokeCommandInfo);
               hWnd := Application.Handle;
               nShow := SW_SHOWNORMAL;
               lpVerb := PChar(MII.wID - CmdID_First);
            end; { with }
            OLECheck(ContextMenu.InvokeCommand(ICI));
            Result := True;
            //Wait for the shell to update its Info files
            SHChangeNotify(SHCNE_UPDATEDIR, SHCNF_FLUSH or SHCNF_IDLIST, TrashPIDL, Nil);
         end; { then }
      finally
         OLECheckBoolean(DestroyMenu(Popup));
      end; { try }
   finally
      SM.Free(TrashPIDL);
   end; { try }
end; { EmptyTrash }

end. { DUTUtilsTrash }
